import os
import string
import sys
from random import randint
from enum import Enum
from difflib import SequenceMatcher
import numpy as np
import pandas as pd
# from PySide2.QtCore import *
# from PySide2.QtGui import *
from PySide2.QtWidgets import *

from GridCal.Gui.ProfilesInput.gui import *
from GridCal.Gui.ProfilesInput.excel_dialog import *


class PandasModel(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with a pandas data frame
    """
    def __init__(self, data, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)
        self._data = np.array(data.values)
        self._cols = data.columns
        self._index = data.index.values
        self.r, self.c = np.shape(self._data)
        self.isDate = False

        if len(self._index) > 0:
            if isinstance(self._index[0], np.datetime64):
                self._index = pd.to_datetime(self._index)
                self.isDate = True

        self.formatter = lambda x: "%.2f" % x

    def rowCount(self, parent=None):
        return self.r

    def columnCount(self, parent=None):
        return self.c

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                # return self.formatter(self._data[index.row(), index.column()])
                return str(self._data[index.row(), index.column()])
        return None

    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._cols[p_int]
            elif orientation == QtCore.Qt.Vertical:
                if self._index is None:
                    return p_int
                else:
                    if self.isDate:
                        return self._index[p_int].strftime('%Y/%m/%d  %H:%M.%S')
                    else:
                        return str(self._index[p_int])
        return None


def get_list_model(iterable):
    """
    get Qt list model from a simple iterable
    :param iterable: 
    :return: List model
    """
    list_model = QtGui.QStandardItemModel()
    if iterable is not None:
        for val in iterable:
            # for the list model
            item = QtGui.QStandardItem(val)
            item.setEditable(False)
            list_model.appendRow(item)
    return list_model


class MultiplierType(Enum):
    Mult = 1,
    Cosfi = 2


class ProfileInputGUI(QtWidgets.QDialog):

    def __init__(self, parent=None, list_of_objects=None, magnitudes=(''), use_native_dialogues=True):
        """

        Args:
            parent:
            list_of_objects: List of objects to which set a profile to
            magnitudes: Property of the objects to which set the pandas DataFrame
            list_of_objects: list ob object to modify
            use_native_dialogues: use the native file selection dialogues?
        """
        QtWidgets.QDialog.__init__(self, parent)
        if list_of_objects is None:
            list_of_objects = list()
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Profiles import dialogue')

        self.project_directory = None

        self.magnitudes = magnitudes

        self.use_native_dialogues = use_native_dialogues

        # results
        self.data = None
        self.time = None
        self.zeroed = None
        self.normalized = False

        # units
        self.units = dict()
        self.units['Y'] = 1e24
        self.units['Z'] = 1e21
        self.units['E'] = 1e18
        self.units['P'] = 1e15
        self.units['T'] = 1e12
        self.units['G'] = 1e9
        self.units['M'] = 1e6
        self.units['k'] = 1e3
        self.units['-'] = 1.0
        self.units['m'] = 1e-3
        self.units['µ'] = 1e-6
        self.units['n'] = 1e-9
        self.units['p'] = 1e-12
        self.units['f'] = 1e-15
        self.units['a'] = 1e-18
        self.units['z'] = 1e-21
        self.units['y'] = 1e-24

        relevant_units = ['T', 'G', 'M', 'k', '-', 'm']
        self.ui.units_combobox.addItems(relevant_units)
        self.ui.units_combobox.setCurrentIndex(2)

        # setup the plot widget
        self.ui.plotwidget.canvas.ax.clear()
        self.ui.plotwidget.canvas.draw()

        # initialize the objectives list
        self.objects = list_of_objects

        # initialize associations
        # self.magnitudes = ['P']
        self.also_reactive_power = False
        # if AlsoReactivePower:
        #     self.magnitudes.append('Q')
        # else:
        # self.ui.setQ_on_cosfi_checkbox.setVisible(False)
        # self.ui.set_cosfi_button.setVisible(False)

        self.associations = list()
        mag = [''] * len(self.magnitudes)
        for elm in list_of_objects:
            self.associations.append([elm] + mag + [1, 1, 1])

        self.P_idx = 0
        self.Q_idx = -1
        self.MULT_idx = 0
        self.SCALE_idx = 0
        self.COSFI_idx = 0

        self.ui.splitter.setStretchFactor(0, 3)
        self.ui.splitter.setStretchFactor(1, 7)

        if len(self.associations) > 0:
            self.display_associations()

        self.original_data_frame = None

        self.profile_names = list()

        # click
        self.ui.open_button.clicked.connect(self.import_profile)
        self.ui.doit_button.clicked.connect(lambda: self.set_multiplier(MultiplierType.Mult))
        self.ui.set_multiplier_button.clicked.connect(lambda: self.set_multiplier(MultiplierType.Mult))
        # self.ui.set_cosfi_button.clicked.connect(lambda: self.set_multiplier(MultiplierType.Cosfi))
        self.ui.autolink_button.clicked.connect(self.auto_link)
        self.ui.rnd_link_pushButton.clicked.connect(self.rnd_link)
        self.ui.assign_to_selection_pushButton.clicked.connect(self.link_to_selection)
        self.ui.assign_to_all_pushButton.clicked.connect(self.link_to_all)
        self.ui.doit_button.clicked.connect(self.do_it)
        self.ui.clear_selection_button.clicked.connect(self.clear_selection)

        # double click
        self.ui.sources_list.doubleClicked.connect(self.sources_list_double_click)
        self.ui.assignation_table.doubleClicked.connect(self.assignation_table_double_click)
        self.ui.tableView.doubleClicked.connect(self.print_profile)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QMessageBox.Ok)
        retval = msg.exec_()

    def get_multiplier(self):
        """
        Gets the necessary multiplier to pass the profile units to Mega
        Remember that the power units in GridCal are the MVA
        """
        unit = self.ui.units_combobox.currentText()
        return self.units[unit] / self.units['M']

    def import_profile(self):
        """
        Select a file to be loaded
        """

        # declare the allowed file types
        files_types = "Formats (*.xlsx *.xls *.csv)"
        # call dialog to select the file
        # filename, type_selected = QFileDialog.getOpenFileNameAndFilter(self, 'Save file', '', files_types)

        # call dialog to select the file

        options = QFileDialog.Options()
        if self.use_native_dialogues:
            options |= QFileDialog.DontUseNativeDialog

        filename, type_selected = QFileDialog.getOpenFileName(self, 'Open file',
                                                              filter=files_types,
                                                              options=options)

        if len(filename) > 0:
            # get the filename extension
            name, file_extension = os.path.splitext(filename)

            # Depending on the extension load the file
            if file_extension == '.csv':
                self.original_data_frame = pd.read_csv(filename, index_col=0)

            elif file_extension in ['.xlsx', '.xls']:

                # select the sheet from the file
                window = ExcelDialog(self, filename)
                window.exec_()
                sheet_index = window.excel_sheet

                if sheet_index is not None:

                    self.original_data_frame = pd.read_excel(filename, sheet_name=sheet_index, index_col=0)

                else:
                    return

            # try to format the data
            try:
                self.original_data_frame = self.original_data_frame.astype(float)
            except:
                self.msg('The format of the data is not recognized. Only int or float values are allowed')
                return

            # set the profile names list
            self.profile_names = np.array([str(e).strip() for e in self.original_data_frame.columns.values], dtype=object)

            # set the loaded data_frame to the GUI
            model = PandasModel(self.original_data_frame)
            self.ui.tableView.setModel(model)

            self.ui.sources_list.setModel(get_list_model(self.original_data_frame.columns))

    def sources_list_double_click(self):
        """
        When an item in the sources list is double clicked, plot the series
        :return:
        """
        if self.original_data_frame is not None:
            idx = self.ui.sources_list.selectedIndexes()[0].row()

            col_name = self.original_data_frame.columns[idx]

            self.ui.plotwidget.canvas.ax.clear()
            self.original_data_frame[col_name].plot(ax=self.ui.plotwidget.canvas.ax)
            self.ui.plotwidget.canvas.draw()

    def display_associations(self):
        """

        @return:
        """
        cols = ['Objective'] + self.magnitudes + ['Scale', 'Cos(φ)', 'Multiplier']
        self.P_idx = 1
        self.Q_idx = self.P_idx + 1
        self.SCALE_idx = len(cols) - 3
        self.COSFI_idx = len(cols) - 2
        self.MULT_idx = len(cols) - 1
        df = pd.DataFrame(self.associations, columns=cols)
        self.ui.assignation_table.setModel(PandasModel(df))

    def print_profile(self):
        """
        prints the profile clicked on the table
        @return:
        """
        if self.original_data_frame is not None:
            idx = self.ui.tableView.selectedIndexes()[0].column()
            name = self.profile_names[idx]
            if idx >= 0:
                self.ui.plotwidget.canvas.ax.clear()
                self.original_data_frame[name].plot(ax=self.ui.plotwidget.canvas.ax)
                self.ui.plotwidget.canvas.draw()

    def objectives_list_double_click(self):
        """
        Link source to objective when the objective item is double clicked
        :return:
        """
        if len(self.ui.sources_list.selectedIndexes()) > 0:
            idx_s = self.ui.sources_list.selectedIndexes()[0].row()
            idx_o = self.ui.objectives_list.selectedIndexes()[0].row()
            scale = self.get_multiplier()
            cosfi = 0.9
            mult = 1
            self.make_association(idx_s, idx_o, scale, cosfi, mult)
            self.display_associations()

    def make_association(self, source_idx, obj_idx, scale=None, cosfi=None, mult=None, col_idx=None):
        """
        Makes an association in the associations table
        """
        col_name = self.original_data_frame.columns[source_idx]

        if scale is None:
            scale = self.get_multiplier()

        if cosfi is None:
            cosfi = 0.9

        if mult is None:
            mult = 1.0

        if col_idx is None:
            col_idx = self.P_idx

        self.associations[obj_idx][col_idx] = self.profile_names[source_idx]
        self.associations[obj_idx][self.SCALE_idx] = scale
        self.associations[obj_idx][self.COSFI_idx] = cosfi
        self.associations[obj_idx][self.MULT_idx] = mult

    def assignation_table_double_click(self):
        """
        Set the selected profile into the clicked slot
        """
        if len(self.ui.sources_list.selectedIndexes()) > 0:
            idx_s = self.ui.sources_list.selectedIndexes()[0].row()
            idx_o = self.ui.assignation_table.selectedIndexes()[0].row()
            col = self.ui.assignation_table.selectedIndexes()[0].column()

            if 0 < col < len(self.associations[idx_o]) - 1:
                self.make_association(idx_s, idx_o, mult=None, col_idx=col)

            self.display_associations()

    def set_multiplier(self, type):
        """
        Set the table multipliers
        """
        if len(self.ui.assignation_table.selectedIndexes()) > 0:
            mult = self.ui.multSpinBox.value()
            for index in self.ui.assignation_table.selectedIndexes():
                idx = index.row()

                if type == MultiplierType.Mult:
                    col = self.MULT_idx
                elif type == MultiplierType.Cosfi:
                    col = self.COSFI_idx
                    if mult > 1 or mult < -1:
                        mult = 0

                self.associations[idx][col] = mult

            self.display_associations()

    @staticmethod
    def normalize_string(s):
        """
        Normalizes a string
        """
        for p in string.punctuation:
            s = s.replace(p, '')
        return s.lower().strip()

    def check_simularity(self, name_to_search, names_array, threshold):
        """

        :param name_to_search:
        :param names_array:
        :param threshold:
        :return:
        """

        if name_to_search in names_array:
            return np.where(names_array == name_to_search)[0][0]

        # else determine the likelihood
        if threshold > 0.01:
            max_val = 0
            max_idx = None
            for idx_s, col_name in enumerate(names_array):
                profile_name = col_name.strip()

                # find the string distance
                d = SequenceMatcher(None, name_to_search, profile_name).ratio()

                if d > max_val:
                    max_val = d
                    max_idx = idx_s

            # assign the string with the closest profile (60% or better similarity)
            if max_idx is not None and max_val > threshold:
                return max_idx
            else:
                return None
        else:
            return None

    def auto_link(self):
        """
        Performs an automatic link between the sources and the objectives based on the names
        """
        mult = self.get_multiplier()
        threshold = self.ui.autolink_slider.value() / 100.0

        for idx_o, elm in enumerate(self.objects):

            idx = self.check_simularity(name_to_search=elm.name.strip(),
                                        names_array=self.profile_names,
                                        threshold=threshold)

            # assign the string with the closest profile (60% or better similarity)
            if idx is not None:
                self.make_association(idx, idx_o, mult)

        self.display_associations()

    def rnd_link(self):
        """
        Random link
        """
        # scale = self.get_multiplier()
        cosfi = 0.9
        mult = 1
        scale = self.get_multiplier()

        if self.ui.sources_list.model() is not None:

            if self.ui.sources_list.model().rowCount() > 0:
                # make a list of the source indices
                source_indices = [i for i in range(self.ui.sources_list.model().rowCount())]

                # make a list of the destination indices
                destination_indices = [i for i in range(self.ui.assignation_table.model().rowCount())]

                # while there are elements in the destination indices
                while len(destination_indices) > 0:

                    # pick a random source
                    rnd_idx_s = randint(0, len(source_indices)-1)

                    # pick and delete a random destination
                    rnd_idx_o = randint(0, len(destination_indices)-1)

                    # get the actual index
                    idx_s = source_indices[rnd_idx_s]

                    # get the actual index
                    idx_o = destination_indices.pop(rnd_idx_o)

                    # make the association
                    self.make_association(idx_s, idx_o, scale, cosfi, mult)

                self.display_associations()
            else:
                pass
        else:
            pass

    def link_to_selection(self):
        """
        Links the selected origin with the selected destinations
        """

        if len(self.ui.sources_list.selectedIndexes()) > 0:
            idx_s = self.ui.sources_list.selectedIndexes()[0].row()

            scale = self.get_multiplier()
            cosfi = 0.9
            mult = 1
            scale = self.get_multiplier()

            # set of different rows
            sel_rows = {item.row() for item in self.ui.assignation_table.selectedIndexes()}

            for idx_o in sel_rows:
                self.make_association(idx_s, idx_o, scale, cosfi, mult)

            self.display_associations()

    def link_to_all(self):
        """
        Links the selected origin with all the destinations
        """

        if len(self.ui.sources_list.selectedIndexes()) > 0:
            idx_s = self.ui.sources_list.selectedIndexes()[0].row()
            cosfi = 0.9
            mult = 1
            scale = self.get_multiplier()

            # set of different rows
            n_rows = self.ui.assignation_table.model().rowCount()

            for idx_o in range(n_rows):
                self.make_association(idx_s, idx_o, scale, cosfi, mult)

            self.display_associations()

    def get_profile(self, parent=None, labels=None, alsoQ=None):
        """
        Return ths assigned profiles
        @return:
            Array of profiles assigned to the input objectives
            Array specifying which objectives are not assigned
        """

        if self.original_data_frame is None:
            return None, None, None

        n_obj = len(self.objects)
        rows_o, cols_o = np.shape(self.original_data_frame)

        profiles = [None] * n_obj
        zeroed = [False] * n_obj  # array to know which profiles are only zeros

        for i_obj in range(n_obj):

            scale = self.associations[i_obj][self.SCALE_idx]
            cosfi = self.associations[i_obj][self.COSFI_idx]
            mult = self.associations[i_obj][self.MULT_idx]
            profile_name = self.associations[i_obj][self.P_idx]

            if profile_name != '':
                # active power
                if self.also_reactive_power:
                    vals = self.original_data_frame[profile_name].values * scale * mult + 0j
                else:
                    vals = self.original_data_frame[profile_name].values * scale * mult

                # add the reactive power if applicable
                if self.also_reactive_power:

                    if self.ui.setQ_on_cosfi_checkbox.isChecked():
                        # fill the data using the power factor
                        fi = np.arccos(cosfi)
                        Q = vals * np.tan(fi)
                    else:
                        # fill Q using the given profile for Q
                        profile_name = self.associations[i_obj][self.Q_idx]

                        if profile_name != '':
                            # fill Q with the set profile
                            Q = self.original_data_frame[profile_name].values * scale * mult
                        else:
                            # if the Q profile name is not given, and is not meant to be made with the power factor
                            # return an array of zeros
                            Q = np.zeros_like(vals)

                    vals += 1j * Q
            else:
                vals = np.zeros(rows_o)
                zeroed[i_obj] = True

            profiles[i_obj] = vals

        time_profile = self.original_data_frame.index

        return np.array(profiles).transpose(), time_profile, zeroed

    def get_association_data(self):
        """
        Return a dictionary with the data association
        @return:
        """
        data = dict()

        if self.original_data_frame is not None:
            for i in range(len(self.associations)):
                objective_name = self.associations[i][0]

                scale = self.associations[i][self.SCALE_idx]
                cosfi = self.associations[i][self.COSFI_idx]
                mult = self.associations[i][self.MULT_idx]

                dta = list()
                for magnitude_idx in range(len(self.magnitudes)):
                    profile_name = self.associations[i][magnitude_idx + 1]

                    if magnitude_idx == self.Q_idx and self.ui.setQ_on_cosfi_checkbox.isChecked():

                        # create Q profile based on P and cosfi
                        profile_name = self.associations[i][self.P_idx]
                        vals = self.original_data_frame[profile_name].values * scale * mult
                        fi = np.arccos(cosfi)
                        vals *= np.tan(fi)
                    else:

                        # pick the profile in the cell
                        if profile_name != '':
                            vals = self.original_data_frame[profile_name].values * scale * mult
                        else:
                            # if there is no profile, output zeros
                            vals = np.zeros(self.original_data_frame.shape[0])

                    dta.append(vals)
                data[objective_name] = np.array(dta).transpose()

        return data

    def clear_selection(self):
        """
        Clear the selected associations
        """
        for idx in self.ui.assignation_table.selectedIndexes():
            obj_idx = idx.row()
            self.associations[obj_idx][self.P_idx] = ""
        self.display_associations()

    def do_it(self):
        """
        Close. The data has to be queried later to the object by the parent by calling get_association_data
        """

        # Generate profiles
        self.data, self.time, self.zeroed = self.get_profile()
        self.normalized = self.ui.normalized_checkBox.isChecked()

        if self.normalized:
            self.data /= self.data.max(axis=0)  # divide each series by the maximum
            self.data = np.nan_to_num(self.data)  # set nan to zero

        self.close()


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    window = ProfileInputGUI(list_of_objects=['Test object'] * 10)
    window.resize(1.61 * 700.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec_())




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import time
import datetime
import numpy as np
from numba import jit, prange
from itertools import combinations
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit
from GridCal.Engine.Simulations.NK.n_minus_k_driver import NMinusKOptions
from GridCal.Engine.Simulations.NK.n_minus_k_ts_results import NMinusKTimeSeriesResults
from GridCal.Engine.Simulations.LinearFactors.analytic_ptdf import LinearAnalysis


@jit(nopython=True, parallel=False)
def compute_flows_numba_t(e, c, nt, OTDF, Flows, rates, overload_count, max_overload, worst_flows):
    """
    Compute OTDF based flows
    :param nt: number of time steps
    :param ne: number of elements
    :param nc: number of failed elements
    :param OTDF: OTDF matrix (element, failed element)
    :param Flows: base flows matrix (time, element)
    :return: Cube of N-1 Flows (time, elements, contingencies)
    """

    for t in range(nt):
        # the formula is: Fn-1(i) = Fbase(i) + OTDF(i,j) * Fbase(j) here i->line, j->failed line
        flow_n_1 = OTDF[e, c] * Flows[t, c] + Flows[t, e]
        flow_n_1_abs = abs(flow_n_1)

        if rates[t, e] > 0:
            rate = flow_n_1_abs / rates[t, e]

            if rate > 1:
                overload_count[e, c] += 1
                if flow_n_1_abs > max_overload[e, c]:
                    max_overload[e, c] = flow_n_1_abs

        if flow_n_1_abs > abs(worst_flows[t, e]):
            worst_flows[t, e] = flow_n_1


@jit(nopython=True, parallel=True)
def compute_flows_numba(e, nt, nc, OTDF, Flows, rates, overload_count, max_overload, worst_flows, paralelize_from=500):
    """
    Compute OTDF based flows
    :param nt: number of time steps
    :param ne: number of elements
    :param nc: number of failed elements
    :param OTDF: OTDF matrix (element, failed element)
    :param Flows: base flows matrix (time, element)
    :return: Cube of N-1 Flows (time, elements, contingencies)
    """

    if nc < paralelize_from:
        for c in range(nc):
            compute_flows_numba_t(e, c, nt, OTDF, Flows, rates, overload_count, max_overload, worst_flows)
    else:
        for c in prange(nc):
            compute_flows_numba_t(e, c, nt, OTDF, Flows, rates, overload_count, max_overload, worst_flows)


class NMinusKTimeSeries(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'N-1 time series'

    def __init__(self, grid: MultiCircuit, options: NMinusKOptions):
        """
        N - k class constructor
        @param grid: MultiCircuit Object
        @param options: N-k options
        @:param pf_options: power flow options
        """
        QThread.__init__(self)

        # Grid to run
        self.grid = grid

        # Options to use
        self.options = options

        # N-K results
        self.results = NMinusKTimeSeriesResults(n=0, ne=0, nc=0,
                                                time_array=(),
                                                bus_names=(),
                                                branch_names=(),
                                                bus_types=())

        # set cancel state
        self.__cancel__ = False

        self.logger = Logger()

        self.elapsed = 0.0

        self.branch_names = list()

    def get_steps(self):
        """
        Get variations list of strings
        """
        if self.results is not None:
            return [v for v in self.branch_names]
        else:
            return list()

    def n_minus_k(self):
        """
        Run N-1 simulation in series
        :return: returns the results
        """

        self.progress_text.emit("Filtering elements by voltage")

        ts_numeric_circuit = compile_time_circuit(self.grid)
        ne = ts_numeric_circuit.nbr
        nc = ts_numeric_circuit.nbr
        nt = len(ts_numeric_circuit.time_array)

        results = NMinusKTimeSeriesResults(ne=ne, nc=nc,
                                           time_array=ts_numeric_circuit.time_array,
                                           n=ts_numeric_circuit.nbus,
                                           branch_names=ts_numeric_circuit.branch_names,
                                           bus_names=ts_numeric_circuit.bus_names,
                                           bus_types=ts_numeric_circuit.bus_types)

        self.progress_text.emit('Analyzing outage distribution factors...')
        linear_analysis = LinearAnalysis(grid=self.grid,
                                         distributed_slack=self.options.distributed_slack,
                                         correct_values=self.options.correct_values)
        linear_analysis.run()

        self.progress_text.emit('Computing branch base flows...')
        Pbus = ts_numeric_circuit.Sbus.real
        flows = linear_analysis.get_branch_time_series(Pbus)
        rates = ts_numeric_circuit.Rates.T

        self.progress_text.emit('Computing N-1 flows...')

        for e in range(ne):
            compute_flows_numba(e=e,
                                nt=nt,
                                nc=nc,
                                OTDF=linear_analysis.results.LODF,
                                Flows=flows,
                                rates=rates,
                                overload_count=results.overload_count,
                                max_overload=results.max_overload,
                                worst_flows=results.worst_flows)

            self.progress_signal.emit((e + 1) / ne * 100)

        results.relative_frequency = results.overload_count / nt
        results.worst_loading = results.worst_flows / (rates + 1e-9)

        return results

    def run(self):
        """

        :return:
        """
        start = time.time()
        self.results = self.n_minus_k()

        end = time.time()
        self.elapsed = end - start
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        self.__cancel__ = True


if __name__ == '__main__':
    import os
    import pandas as pd
    from GridCal.Engine import FileOpen, SolverType, PowerFlowOptions

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/2869 Pegase.gridcal'
    # fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', 'IEEE 30 Bus with storage.xlsx')
    # fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', '2869 Pegase.gridcal')

    main_circuit = FileOpen(fname).open()

    options_ = NMinusKOptions()
    simulation = NMinusKTimeSeries(grid=main_circuit, options=options_)
    simulation.run()

    print()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import json
import pandas as pd
import numpy as np
import time
import multiprocessing
from sklearn.cluster import KMeans
from PySide2.QtCore import QThread, QThreadPool, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import single_island_pf, power_flow_worker_args
from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit, BranchImpedanceMode
from GridCal.Engine.Simulations.Stochastic.latin_hypercube_sampling import lhs
from GridCal.Engine.Simulations.results_model import ResultsModel


class TimeSeriesResults(PowerFlowResults):

    def __init__(self, n, m, n_tr, n_hvdc, bus_names, branch_names, transformer_names, hvdc_names,
                 time_array, bus_types):
        """
        TimeSeriesResults constructor
        :param n: number of buses
        :param m: number of branches
        :param n_tr:
        :param n_hvdc:
        :param bus_names:
        :param branch_names:
        :param transformer_names:
        :param hvdc_names:
        :param time_array:
        :param bus_types:
        """
        PowerFlowResults.__init__(self,
                                  n=n,
                                  m=m,
                                  n_tr=n_tr,
                                  n_hvdc=n_hvdc,
                                  bus_names=bus_names,
                                  branch_names=branch_names,
                                  transformer_names=transformer_names,
                                  hvdc_names=hvdc_names,
                                  bus_types=bus_types)

        self.name = 'Time series'
        self.nt = len(time_array)
        self.m = m
        self.n = n

        self.time = time_array

        self.bus_types = np.zeros(n, dtype=int)

        self.voltage = np.zeros((self.nt, n), dtype=complex)

        self.S = np.zeros((self.nt, n), dtype=complex)

        self.Sf = np.zeros((self.nt, m), dtype=complex)

        self.St = np.zeros((self.nt, m), dtype=complex)

        self.Vbranch = np.zeros((self.nt, m), dtype=complex)

        self.loading = np.zeros((self.nt, m), dtype=complex)

        self.losses = np.zeros((self.nt, m), dtype=complex)

        self.hvdc_losses = np.zeros((self.nt, self.n_hvdc))

        self.hvdc_Pf = np.zeros((self.nt, self.n_hvdc))

        self.hvdc_Pt = np.zeros((self.nt, self.n_hvdc))

        self.hvdc_loading = np.zeros((self.nt, self.n_hvdc))

        self.flow_direction = np.zeros((self.nt, m), dtype=float)

        self.error_values = np.zeros(self.nt)

        self.converged_values = np.ones(self.nt, dtype=bool)  # guilty assumption

        self.overloads = [None] * self.nt

        self.overvoltage = [None] * self.nt

        self.undervoltage = [None] * self.nt

        self.overloads_idx = [None] * self.nt

        self.overvoltage_idx = [None] * self.nt

        self.undervoltage_idx = [None] * self.nt

        self.buses_useful_for_storage = [None] * self.nt

        # results available
        self.available_results = [ResultTypes.BusVoltageModule,
                                  ResultTypes.BusVoltageAngle,
                                  ResultTypes.BusActivePower,
                                  ResultTypes.BusReactivePower,

                                  ResultTypes.BranchActivePowerFrom,
                                  ResultTypes.BranchReactivePowerFrom,
                                  ResultTypes.BranchLoading,
                                  ResultTypes.BranchActiveLosses,
                                  ResultTypes.BranchReactiveLosses,
                                  ResultTypes.BranchVoltage,
                                  ResultTypes.BranchAngles,
                                  ResultTypes.SimulationError,

                                  ResultTypes.HvdcLosses,
                                  ResultTypes.HvdcPowerFrom,
                                  ResultTypes.HvdcPowerTo]

    def set_at(self, t, results: PowerFlowResults):
        """
        Set the results at the step t
        @param t: time index
        @param results: PowerFlowResults instance
        """

        self.voltage[t, :] = results.voltage

        self.S[t, :] = results.Sbus

        self.Sf[t, :] = results.Sf
        self.St[t, :] = results.St

        self.Vbranch[t, :] = results.Vbranch

        self.loading[t, :] = results.loading

        self.losses[t, :] = results.losses

        self.flow_direction[t, :] = results.flow_direction

        self.error_values[t] = results.error

        self.converged_values[t] = results.converged

    @staticmethod
    def merge_if(df, arr, ind, cols):
        """

        @param df:
        @param arr:
        @param ind:
        @param cols:
        @return:
        """
        obj = pd.DataFrame(data=arr, index=ind, columns=cols)
        if df is None:
            df = obj
        else:
            df = pd.concat([df, obj], axis=1)

        return df

    def apply_from_island(self, results, b_idx, br_idx, t_index, grid_idx):
        """
        Apply results from another island circuit to the circuit results represented here
        :param results: PowerFlowResults
        :param b_idx: bus original indices
        :param br_idx: branch original indices
        :param t_index:
        :param grid_idx:
        :return:
        """

        # bus results
        if self.voltage.shape == results.voltage.shape:
            self.voltage = results.voltage
            self.S = results.S
        else:
            self.voltage[np.ix_(t_index, b_idx)] = results.voltage
            self.S[np.ix_(t_index, b_idx)] = results.S

        # branch results
        if self.Sf.shape == results.Sf.shape:
            self.Sf = results.Sf
            self.St = results.St

            self.Vbranch = results.Vbranch

            self.loading = results.loading

            self.losses = results.losses

            self.flow_direction = results.flow_direction

            if (results.error_values > self.error_values).any():
                self.error_values += results.error_values

            self.converged_values = self.converged_values * results.converged_values

        else:
            self.Sf[np.ix_(t_index, br_idx)] = results.Sf
            self.St[np.ix_(t_index, br_idx)] = results.St

            self.Vbranch[np.ix_(t_index, br_idx)] = results.Vbranch

            self.loading[np.ix_(t_index, br_idx)] = results.loading

            self.losses[np.ix_(t_index, br_idx)] = results.losses

            self.flow_direction[np.ix_(t_index, br_idx)] = results.flow_direction

            if (results.error_values > self.error_values[t_index]).any():
                self.error_values[t_index] += results.error_values

            self.converged_values[t_index] = self.converged_values[t_index] * results.converged_values

    def get_results_dict(self):
        """
        Returns a dictionary with the results sorted in a dictionary
        :return: dictionary of 2D numpy arrays (probably of complex numbers)
        """
        data = {'Vm': np.abs(self.voltage).tolist(),
                'Va': np.angle(self.voltage).tolist(),
                'P': self.S.real.tolist(),
                'Q': self.S.imag.tolist(),
                'Sf_real': self.Sf.real.tolist(),
                'Sf_imag': self.Sf.imag.tolist(),
                'loading': np.abs(self.loading).tolist(),
                'losses_real': np.real(self.losses).tolist(),
                'losses_imag': np.imag(self.losses).tolist()}
        return data

    def save(self, fname):
        """
        Export as json
        """

        with open(fname, "wb") as output_file:
            json_str = json.dumps(self.get_results_dict())
            output_file.write(json_str)

    def analyze(self):
        """
        Analyze the results
        @return:
        """
        branch_overload_frequency = np.zeros(self.m)
        bus_undervoltage_frequency = np.zeros(self.n)
        bus_overvoltage_frequency = np.zeros(self.n)
        buses_selected_for_storage_frequency = np.zeros(self.n)
        for i in range(self.nt):
            branch_overload_frequency[self.overloads_idx[i]] += 1
            bus_undervoltage_frequency[self.undervoltage_idx[i]] += 1
            bus_overvoltage_frequency[self.overvoltage_idx[i]] += 1
            buses_selected_for_storage_frequency[self.buses_useful_for_storage[i]] += 1

        return branch_overload_frequency, \
                bus_undervoltage_frequency, \
                bus_overvoltage_frequency, \
                buses_selected_for_storage_frequency

    def mdl(self, result_type: ResultTypes) -> "ResultsModel":
        """

        :param result_type:
        :return:
        """

        if result_type == ResultTypes.BusVoltageModule:
            labels = self.bus_names
            data = np.abs(self.voltage)
            y_label = '(p.u.)'
            title = 'Bus voltage '

        elif result_type == ResultTypes.BusVoltageAngle:
            labels = self.bus_names
            data = np.angle(self.voltage, deg=True)
            y_label = '(Deg)'
            title = 'Bus voltage '

        elif result_type == ResultTypes.BusActivePower:
            labels = self.bus_names
            data = self.S.real
            y_label = '(MW)'
            title = 'Bus active power '

        elif result_type == ResultTypes.BusReactivePower:
            labels = self.bus_names
            data = self.S.imag
            y_label = '(MVAr)'
            title = 'Bus reactive power '

        elif result_type == ResultTypes.BranchPower:
            labels = self.branch_names
            data = self.Sf
            y_label = '(MVA)'
            title = 'Branch power '

        elif result_type == ResultTypes.BranchActivePowerFrom:
            labels = self.branch_names
            data = self.Sf.real
            y_label = '(MW)'
            title = 'Branch power '

        elif result_type == ResultTypes.BranchReactivePowerFrom:
            labels = self.branch_names
            data = self.Sf.imag
            y_label = '(MVAr)'
            title = 'Branch power '

        elif result_type == ResultTypes.BranchLoading:
            labels = self.branch_names
            data = np.abs(self.loading) * 100
            y_label = '(%)'
            title = 'Branch loading '

        elif result_type == ResultTypes.BranchLosses:
            labels = self.branch_names
            data = self.losses
            y_label = '(MVA)'
            title = 'Branch losses'

        elif result_type == ResultTypes.BranchActiveLosses:
            labels = self.branch_names
            data = self.losses.real
            y_label = '(MW)'
            title = 'Branch losses'

        elif result_type == ResultTypes.BranchReactiveLosses:
            labels = self.branch_names
            data = self.losses.imag
            y_label = '(MVAr)'
            title = 'Branch losses'

        elif result_type == ResultTypes.BranchVoltage:
            labels = self.branch_names
            data = np.abs(self.Vbranch)
            y_label = '(p.u.)'
            title = result_type.value[0]

        elif result_type == ResultTypes.BranchAngles:
            labels = self.branch_names
            data = np.angle(self.Vbranch, deg=True)
            y_label = '(deg)'
            title = result_type.value[0]

        elif result_type == ResultTypes.BatteryPower:
            labels = self.branch_names
            data = np.zeros_like(self.losses)
            y_label = '$\Delta$ (MVA)'
            title = 'Battery power'

        elif result_type == ResultTypes.SimulationError:
            data = self.error_values.reshape(-1, 1)
            y_label = 'p.u.'
            labels = ['Error']
            title = 'Error'

        elif result_type == ResultTypes.HvdcLosses:
            labels = self.hvdc_names
            data = self.hvdc_losses
            y_label = '(MW)'
            title = result_type.value

        elif result_type == ResultTypes.HvdcPowerFrom:
            labels = self.hvdc_names
            data = self.hvdc_Pf
            y_label = '(MW)'
            title = result_type.value

        elif result_type == ResultTypes.HvdcPowerTo:
            labels = self.hvdc_names
            data = self.hvdc_Pt
            y_label = '(MW)'
            title = result_type.value

        else:
            raise Exception('Result type not understood:' + str(result_type))

        if self.time is not None:
            index = self.time
        else:
            index = list(range(data.shape[0]))

        # assemble model
        mdl = ResultsModel(data=data, index=index, columns=labels, title=title, ylabel=y_label, units=y_label)
        return mdl


def kmeans_case_sampling(X, n_points=10):
    """
    K-Means clustering
    :param X: injections matrix (time, bus)
    :param n_points: number of clusters
    :return: indices of the closest to the cluster centers, deviation of the closest representatives
    """

    # declare the model
    model = KMeans(n_clusters=n_points)

    # model fitting
    model.fit(X)

    centers = model.cluster_centers_
    labels = model.labels_

    # get the closest indices to the cluster centers
    closest_idx = np.zeros(n_points, dtype=int)
    closest_prob = np.zeros(n_points, dtype=float)
    nt = X.shape[0]

    unique_labels, counts = np.unique(labels, return_counts=True)
    probabilities = counts.astype(float) / float(nt)

    prob_dict = {u: p for u, p in zip(unique_labels, probabilities)}
    for i in range(n_points):
        deviations = np.sum(np.power(X - centers[i, :], 2.0), axis=1)
        idx = deviations.argmin()
        closest_idx[i] = idx

    # sort the indices
    closest_idx = np.sort(closest_idx)

    # compute the probabilities of each index (sorted already)
    for i, idx in enumerate(closest_idx):
        lbl = model.predict(X[idx, :].reshape(1, -1))[0]
        prob = prob_dict[lbl]
        closest_prob[i] = prob

    return closest_idx, closest_prob


def time_series_worker(n, m, time_profile, namespace, options: PowerFlowOptions,
                       time_indices, logger: Logger) -> (TimeSeriesResults, np.array):
    """

    :param n:
    :param m:
    :param time_profile:
    :param options:
    :param time_indices:  array of time indices to consider
    :param logger:
    :return: TimeSeriesResults instance
    """

    # initialize the grid time series results we will append the island results with another function
    time_series_results = TimeSeriesResults(n, m, time_array=time_profile[time_indices])

    calc_inputs_dict = namespace.calc_inputs_dict
    time_series_results.bus_types = namespace.bus_types

    # for each partition of the profiles...
    for t_key, calc_inputs in calc_inputs_dict.items():

        # For every island, run the time series
        for island_index, calculation_input in enumerate(calc_inputs):

            # find the original indices
            bus_original_idx = calculation_input.original_bus_idx
            branch_original_idx = calculation_input.original_branch_idx

            # if there are valid profiles...
            if time_profile is not None:

                # declare a results object for the partition
                results = TimeSeriesResults(n=calculation_input.nbus,
                                            m=calculation_input.nbr,
                                            time_array=time_profile[time_indices])
                last_voltage = calculation_input.Vbus

                # default value in case of single-valued profile
                dt = 1.0

                # traverse the time profiles of the partition and simulate each time step
                for it, t in enumerate(time_indices):

                    # set the power values
                    # if the storage dispatch option is active, the batteries power is not included
                    # therefore, it shall be included after processing
                    Ysh = calculation_input.Ysh_prof[:, it]
                    I = calculation_input.Ibus_prof[:, it]
                    S = calculation_input.Sbus_prof[:, it]
                    branch_rates = calculation_input.branch_rates_prof[it, :]

                    # add the controlled storage power if we are controlling the storage devices
                    if options.dispatch_storage:

                        if (it+1) < len(calculation_input.original_time_idx):
                            # compute the time delta: the time values come in nanoseconds
                            dt = (calculation_input.time_array[it + 1]
                                  - calculation_input.time_array[it]).value * 1e-9 / 3600.0

                    # run power flow at the circuit
                    res = single_island_pf(circuit=calculation_input,
                                           Vbus=last_voltage,
                                           Sbus=S,
                                           Ibus=I,
                                           branch_rates=branch_rates,
                                           options=options,
                                           logger=logger)

                    # store circuit results at the time index 'it'
                    results.set_at(it, res)

                # merge the circuit's results
                time_series_results.apply_from_island(results,
                                                      bus_original_idx,
                                                      branch_original_idx,
                                                      time_indices,
                                                      'TS')
    return time_series_results, time_indices


class TimeSeries(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Time Series'

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions, opf_time_series_results=None,
                 start_=0, end_=None, use_clustering=False, cluster_number=10):
        """
        TimeSeries constructor
        @param grid: MultiCircuit instance
        @param options: PowerFlowOptions instance
        """
        QThread.__init__(self)

        # reference the grid directly
        self.grid = grid

        self.options = options

        self.opf_time_series_results = opf_time_series_results

        self.results = None

        self.start_ = start_

        self.end_ = end_

        self.use_clustering = use_clustering

        self.cluster_number = cluster_number

        self.elapsed = 0

        self.logger = Logger()

        self.returned_results = list()

        self.pool = None

        self._mt_i = 0
        self._mt_n = 1

        self.__cancel__ = False

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return [l.strftime('%d-%m-%Y %H:%M') for l in pd.to_datetime(self.grid.time_profile[self.start_: self.end_])]

    def run_single_thread(self, time_indices) -> TimeSeriesResults:
        """
        Run single thread time series
        :param time_indices: array of time indices to consider
        :return: TimeSeriesResults instance
        """

        # compile the multi-circuit
        time_circuit = compile_time_circuit(circuit=self.grid,
                                            apply_temperature=False,
                                            branch_tolerance_mode=BranchImpedanceMode.Specified,
                                            opf_results=self.opf_time_series_results)

        # do the topological computation
        time_islands = time_circuit.split_into_islands(ignore_single_node_islands=self.options.ignore_single_node_islands)

        # initialize the grid time series results we will append the island results with another function
        time_series_results = TimeSeriesResults(n=time_circuit.nbus,
                                                m=time_circuit.nbr,
                                                n_tr=time_circuit.ntr,
                                                n_hvdc=time_circuit.nhvdc,
                                                bus_names=time_circuit.bus_names,
                                                branch_names=time_circuit.branch_names,
                                                transformer_names=time_circuit.tr_names,
                                                hvdc_names=time_circuit.hvdc_names,
                                                bus_types=time_circuit.bus_types,
                                                time_array=self.grid.time_profile[time_indices])

        time_series_results.bus_types = time_circuit.bus_types

        # For every island, run the time series
        for island_index, calculation_input in enumerate(time_islands):

            # fill in Vbus, Sbus Ibus
            # calculation_input.consolidate()

            # Are we dispatching storage? if so, generate a dictionary of battery -> bus index
            # to be able to set the batteries values into the vector S
            batteries = list()
            batteries_bus_idx = list()
            if self.options.dispatch_storage:
                for k, bus in enumerate(self.grid.buses):
                    for battery in bus.batteries:
                        battery.reset()  # reset the calculation values
                        batteries.append(battery)
                        batteries_bus_idx.append(k)

            self.progress_text.emit('Time series at circuit ' + str(island_index) + '...')

            # find the original indices
            bus_original_idx = calculation_input.original_bus_idx
            branch_original_idx = calculation_input.original_branch_idx

            # declare a results object for the partition
            results = TimeSeriesResults(n=calculation_input.nbus,
                                        m=calculation_input.nbr,
                                        n_tr=calculation_input.ntr,
                                        n_hvdc=calculation_input.nhvdc,
                                        bus_names=calculation_input.bus_names,
                                        branch_names=calculation_input.branch_names,
                                        transformer_names=calculation_input.tr_names,
                                        hvdc_names=calculation_input.hvdc_names,
                                        bus_types=time_circuit.bus_types,
                                        time_array=self.grid.time_profile[time_indices])

            self.progress_signal.emit(0.0)

            # default value in case of single-valued profile
            dt = 1.0

            # traverse the time profiles of the partition and simulate each time step
            for it, t in enumerate(time_indices):

                # set the power values
                # if the storage dispatch option is active, the batteries power is not included
                # therefore, it shall be included after processing
                V = calculation_input.Vbus[:, it]
                # Ysh = calculation_input.Yshunt_from_devices[:, it]
                I = calculation_input.Ibus[:, it]
                S = calculation_input.Sbus[:, it]
                branch_rates = calculation_input.Rates[:, it]

                # add the controlled storage power if we are controlling the storage devices
                if self.options.dispatch_storage:

                    if (it+1) < len(calculation_input.original_time_idx):
                        # compute the time delta: the time values come in nanoseconds
                        dt = (calculation_input.time_array[it + 1]
                              - calculation_input.time_array[it]).value * 1e-9 / 3600.0

                    for k, battery in enumerate(batteries):

                        power = battery.get_processed_at(it, dt=dt, store_values=True)

                        bus_idx = batteries_bus_idx[k]

                        S[bus_idx] += power / calculation_input.Sbase

                # run power flow at the circuit
                res = single_island_pf(circuit=calculation_input,
                                       Vbus=V,
                                       Sbus=S,
                                       Ibus=I,
                                       branch_rates=branch_rates,
                                       options=self.options,
                                       logger=self.logger)

                # Recycle voltage solution
                # last_voltage = res.voltage

                # store circuit results at the time index 'it'
                results.set_at(it, res)

                progress = ((t - self.start_ + 1) / (self.end_ - self.start_)) * 100
                self.progress_signal.emit(progress)
                self.progress_text.emit('Simulating island ' + str(island_index)
                                        + ' at ' + str(self.grid.time_profile[t]))

                if self.__cancel__:
                    # merge the circuit's results
                    time_series_results.apply_from_island(results,
                                                          bus_original_idx,
                                                          branch_original_idx,
                                                          time_indices,
                                                          'TS')
                    # abort by returning at this point
                    return time_series_results

            # merge the circuit's results
            time_series_results.apply_from_island(results,
                                                  bus_original_idx,
                                                  branch_original_idx,
                                                  time_indices,
                                                  'TS')

        # set the HVDC results here since the HVDC is not a branch in this modality
        time_series_results.hvdc_Pf = -time_circuit.hvdc_Pf.T
        time_series_results.hvdc_Pt = -time_circuit.hvdc_Pt.T
        time_series_results.hvdc_loading = time_circuit.hvdc_loading.T
        time_series_results.hvdc_losses = time_circuit.hvdc_losses.T

        return time_series_results

    def run_single_thread_clustering(self, time_indices) -> TimeSeriesResults:
        """
        Run single thread time series using the time series clustering
        :param time_indices: array of time indices to consider
        :return: TimeSeriesResults instance
        """
        # compile the multi-circuit
        numerical_circuit = self.grid.compile_time_series(opf_time_series_results=self.opf_time_series_results)

        self.progress_text.emit('Clustering...')
        X = numerical_circuit.get_power_injections()
        X = X[:, time_indices].real.T
        time_idx, closest_prob = kmeans_case_sampling(X, n_points=self.cluster_number)

        time_series_results = self.run_single_thread(time_indices=time_idx)

        return time_series_results

    def update_prog(self):
        self._mt_i += 1
        progress = (self._mt_i + 1) / self._mt_n * 100
        self.progress_signal.emit(progress)

    def collect_mt_result(self, res):
        """

        :param res:
        :return:
        """
        self.returned_results.append(res)

    def run_multi_thread_old(self, time_indices) -> TimeSeriesResults:
        """
        Run multi thread time series
        :param time_indices: array of time indices to consider
        :return: TimeSeriesResults instance
        """

        # initialize the grid time series results, we will append the island results with another function
        n = len(self.grid.buses)
        m = self.grid.get_branch_number()
        time_series_results = TimeSeriesResults(n, m, time_array=self.grid.time_profile[time_indices])

        n_cores = multiprocessing.cpu_count()
        self.pool = multiprocessing.Pool()

        # compile the multi-circuit
        numerical_circuit = self.grid.compile_time_series(opf_time_series_results=self.opf_time_series_results)

        # perform the topological computation
        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        # for each partition of the profiles...
        for t_key, calc_inputs in calc_inputs_dict.items():

            # For every island, run the time series
            for island_index, calculation_input in enumerate(calc_inputs):

                self.progress_text.emit('Time series at circuit ' + str(island_index) + '...')

                # Start jobs
                self.returned_results = list()

                # if there are valid profiles...
                if self.grid.time_profile is not None:

                    # declare a results object for the partition
                    results = TimeSeriesResults(n=calculation_input.nbus,
                                                m=calculation_input.nbr,
                                                time_array=self.grid.time_profile[time_indices])
                    Vbus = calculation_input.Vbus

                    self.progress_signal.emit(0.0)

                    # traverse the time profiles of the partition and simulate each time step
                    for it, t in enumerate(time_indices):

                        # set the power values
                        # if the storage dispatch option is active, the batteries power is not included
                        # therefore, it shall be included after processing
                        Ysh = calculation_input.Ysh_prof[:, it]
                        Ibus = calculation_input.Ibus_prof[:, it]
                        Sbus = calculation_input.Sbus_prof[:, it]
                        branch_rates = calculation_input.branch_rates_prof[t, :]

                        args = (it, self.options, calculation_input, Vbus, Sbus, Ibus, branch_rates)
                        self.pool.apply_async(power_flow_worker_args, (args,), callback=self.collect_mt_result)

                    # wait for all jobs to complete
                    self.pool.close()
                    self.pool.join()

                    # collect results
                    self.progress_text.emit('Collecting results...')
                    for it, res in self.returned_results:
                        # store circuit results at the time index 't'
                        results.set_at(it, res)

                    # merge  the circuit's results
                    time_series_results.apply_from_island(results,
                                                          calculation_input.original_bus_idx,
                                                          calculation_input.original_branch_idx,
                                                          time_indices,
                                                          'TS multi-thread')

                else:
                    print('There are no profiles')
                    self.progress_text.emit('There are no profiles')

        return time_series_results

    def run_multi_thread(self, time_indices) -> TimeSeriesResults:

        # initialize the grid time series results, we will append the island results with another function
        n = len(self.grid.buses)
        m = self.grid.get_branch_number()
        nt = len(time_indices)
        time_series_results = TimeSeriesResults(n, m, time_array=self.grid.time_profile[time_indices])

        # compile the multi-circuit
        self.progress_text.emit('Compiling time series...')
        numerical_circuit = self.grid.compile_time_series(opf_time_series_results=self.opf_time_series_results)

        # compute the time indices per process
        n_cores = multiprocessing.cpu_count()
        nchunk = int(nt / n_cores)
        time_chunks = list()
        k = 0
        while k < (nt - nchunk * 2):
            k2 = k+nchunk
            if k2 > nt:
                k2 = nt
            time_chunk = time_indices[k:k2]
            k += nchunk
            time_chunks.append(time_chunk)
        time_chunks.append(time_indices[k::])  # add the last chunk

        # schedule jobs
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Running in parallel...')
        self.pool = multiprocessing.Pool()
        manager = multiprocessing.Manager()
        namespace = manager.Namespace()

        stuff = list()

        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        namespace.calc_inputs_dict = calc_inputs_dict
        namespace.bus_types = numerical_circuit.bus_types

        for time_chunk in time_chunks:
            # n, m, time_profile, buses, numerical_circuit, options, time_indices, logger
            args = (n, m, self.grid.time_profile, namespace, self.options, time_chunk, self.logger)
            p = self.pool.apply_async(func=time_series_worker, args=args, callback=self.collect_mt_result)
            stuff.append(p)

        # wait for all jobs to complete
        self.pool.close()
        self.pool.join()

        # collect results
        self.progress_text.emit('Collecting results...')
        for result, time_chunk in self.returned_results:
            # merge  the circuit's results
            time_series_results.apply_from_island(result,
                                                  np.arange(n),
                                                  np.arange(m),
                                                  time_chunk,
                                                  'TS multi-thread')
        self.progress_signal.emit(100.0)
        return time_series_results

    def run(self):
        """
        Run the time series simulation
        @return:
        """

        a = time.time()

        if self.end_ is None:
            self.end_ = len(self.grid.time_profile)
        time_indices = np.arange(self.start_, self.end_)

        if self.options.multi_thread:
            self.results = self.run_multi_thread(time_indices)
        else:
            if self.use_clustering:
                self.results = self.run_single_thread_clustering(time_indices)
            else:
                self.results = self.run_single_thread(time_indices)

        self.elapsed = time.time() - a

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        if self.pool is not None:
            self.pool.terminate()
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled!')
        self.done_signal.emit()


class SampledTimeSeries(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Time Series'

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions, opf_time_series_results=None,
                 number_of_steps=10):
        """
        TimeSeries constructor
        @param grid: MultiCircuit instance
        @param options: PowerFlowOptions instance
        """
        QThread.__init__(self)

        # reference the grid directly
        self.grid = grid

        self.options = options

        self.opf_time_series_results = opf_time_series_results

        self.results = None

        self.number_of_steps = number_of_steps

        self.elapsed = 0

        self.logger = Logger()

        self.returned_results = list()

        self.pool = None

        self.steps = self.sample_steps()

        self.__cancel__ = False

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return [l.strftime('%d-%m-%Y %H:%M') for l in pd.to_datetime(self.grid.time_profile[self.steps])]

    def sample_steps(self):

        n = len(self.grid.time_profile)
        points = (np.sort(lhs(self.number_of_steps, 1)[0]) * n).astype(int)

        return points

    def run_single_thread(self) -> TimeSeriesResults:
        """
        Run single thread time series
        :return: TimeSeriesResults instance
        """

        # initialize the grid time series results we will append the island results with another function
        n = len(self.grid.buses)
        m = self.grid.get_branch_number()
        nt = self.number_of_steps
        time_series_results = TimeSeriesResults(n, m, time_array=self.grid.time_profile[self.steps])

        # compile the multi-circuit
        numerical_circuit = self.grid.compile_time_series(opf_time_series_results=self.opf_time_series_results)

        # do the topological computation
        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        time_series_results.bus_types = numerical_circuit.bus_types

        # for each partition of the profiles...
        for t_key, calc_inputs in calc_inputs_dict.items():

            # For every island, run the time series
            for island_index, calculation_input in enumerate(calc_inputs):

                # Are we dispatching storage? if so, generate a dictionary of battery -> bus index
                # to be able to set the batteries values into the vector S
                batteries = list()
                batteries_bus_idx = list()
                if self.options.dispatch_storage:
                    for k, bus in enumerate(self.grid.buses):
                        for battery in bus.batteries:
                            battery.reset()  # reset the calculation values
                            batteries.append(battery)
                            batteries_bus_idx.append(k)

                self.progress_text.emit('Time series at circuit ' + str(island_index) + '...')

                # find the original indices
                bus_original_idx = calculation_input.original_bus_idx
                branch_original_idx = calculation_input.original_branch_idx

                # if there are valid profiles...
                if self.grid.time_profile is not None:

                    # declare a results object for the partition
                    # nt = calculation_input.ntime
                    n = calculation_input.nbus
                    m = calculation_input.nbr
                    results = TimeSeriesResults(n, m, nt, 0, nt)
                    last_voltage = calculation_input.Vbus

                    self.progress_signal.emit(0.0)

                    # default value in case of single-valued profile
                    dt = 1.0

                    # traverse the time profiles of the partition and simulate each time step
                    for it, t in enumerate(self.steps):

                        # set the power values
                        # if the storage dispatch option is active, the batteries power is not included
                        # therefore, it shall be included after processing
                        Ysh = calculation_input.Ysh_prof[:, t]
                        I = calculation_input.Ibus_prof[:, t]
                        S = calculation_input.Sbus_prof[:, t]
                        branch_rates = calculation_input.branch_rates_prof[t, :]

                        # add the controlled storage power if we are controlling the storage devices
                        if self.options.dispatch_storage:

                            if (it+1) < len(calculation_input.original_time_idx):
                                # compute the time delta: the time values come in nanoseconds
                                dt = (calculation_input.time_array[it + 1]
                                      - calculation_input.time_array[it]).value * 1e-9 / 3600.0

                            for k, battery in enumerate(batteries):

                                power = battery.get_processed_at(it, dt=dt, store_values=True)

                                bus_idx = batteries_bus_idx[k]

                                S[bus_idx] += power / calculation_input.Sbase
                        else:
                            pass

                        # run power flow at the circuit
                        res = single_island_pf(circuit=calculation_input,
                                               Vbus=last_voltage,
                                               Sbus=S,
                                               Ibus=I,
                                               branch_rates=branch_rates,
                                               options=self.options,
                                               logger=self.logger)

                        # Recycle voltage solution
                        # last_voltage = res.voltage

                        # store circuit results at the time index 't'
                        results.set_at(it, res)

                        progress = ((it + 1) / self.number_of_steps) * 100
                        self.progress_signal.emit(progress)
                        self.progress_text.emit('Simulating island ' + str(island_index)
                                                + ' at ' + str(self.grid.time_profile[t]))

                    if self.__cancel__:
                        # merge the circuit's results
                        time_series_results.apply_from_island(results,
                                                              bus_original_idx,
                                                              branch_original_idx,
                                                              np.array(range(self.number_of_steps)), #np.array(calculation_input.original_time_idx)[self.steps],
                                                              'TS')
                        # abort by returning at this point
                        return time_series_results

                    # merge the circuit's results
                    time_series_results.apply_from_island(results,
                                                          bus_original_idx,
                                                          branch_original_idx,
                                                          np.array(range(self.number_of_steps)),
                                                          'TS')

                else:
                    print('There are no profiles')
                    self.progress_text.emit('There are no profiles')

        return time_series_results

    def update_progress_mt(self, res):
        """

        :param res:
        :return:
        """
        t, _ = res
        progress = ((t - self.start_ + 1) / (self.end_ - self.start_)) * 100
        self.progress_signal.emit(progress)
        self.returned_results.append(res)

    def run_multi_thread(self) -> TimeSeriesResults:
        """
        Run multi thread time series
        :return: TimeSeriesResults instance
        """

        # initialize the grid time series results, we will append the island results with another function
        n = len(self.grid.buses)
        m = self.grid.get_branch_number()
        nt = len(self.grid.time_profile)
        time_series_results = TimeSeriesResults(n, m, nt, time_array=self.grid.time_profile)

        if self.end_ is None:
            self.end_ = nt

        n_cores = multiprocessing.cpu_count()
        self.pool = multiprocessing.Pool()

        # compile the multi-circuit
        numerical_circuit = self.grid.compile_time_series(opf_time_series_results=self.opf_time_series_results)

        # perform the topological computation
        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        # for each partition of the profiles...
        for t_key, calc_inputs in calc_inputs_dict.items():

            # For every island, run the time series
            for island_index, calculation_input in enumerate(calc_inputs):

                self.progress_text.emit('Time series at circuit ' + str(island_index) + '...')

                # Start jobs
                self.returned_results = list()

                # if there are valid profiles...
                if self.grid.time_profile is not None:

                    # declare a results object for the partition
                    nt = calculation_input.ntime
                    n = calculation_input.nbus
                    m = calculation_input.nbr
                    results = TimeSeriesResults(n, m, nt, self.start_, self.end_)
                    Vbus = calculation_input.Vbus

                    self.progress_signal.emit(0.0)

                    # traverse the time profiles of the partition and simulate each time step
                    for it, t in enumerate(calculation_input.original_time_idx):

                        if (t >= self.start_) and (t < self.end_):

                            # set the power values
                            # if the storage dispatch option is active, the batteries power is not included
                            # therefore, it shall be included after processing
                            Ysh = calculation_input.Ysh_prof[:, it]
                            Ibus = calculation_input.Ibus_prof[:, it]
                            Sbus = calculation_input.Sbus_prof[:, it]
                            branch_rates = calculation_input.branch_rates_prof[t, :]

                            args = (t, self.options, calculation_input, Vbus, Sbus, Ibus, branch_rates)
                            self.pool.apply_async(power_flow_worker_args, (args,), callback=self.update_progress_mt)

                    # wait for all jobs to complete
                    self.pool.close()
                    self.pool.join()

                    # collect results
                    self.progress_text.emit('Collecting results...')
                    for t, res in self.returned_results:
                        # store circuit results at the time index 't'
                        results.set_at(t, res)

                    # merge  the circuit's results
                    time_series_results.apply_from_island(results,
                                                          calculation_input.original_bus_idx,
                                                          calculation_input.original_branch_idx,
                                                          calculation_input.original_time_idx,
                                                          'TS multi-thread')

                else:
                    print('There are no profiles')
                    self.progress_text.emit('There are no profiles')

        return time_series_results

    def run(self):
        """
        Run the time series simulation
        @return:
        """

        a = time.time()
        if self.options.multi_thread:
            self.results = self.run_multi_thread()
        else:
            self.results = self.run_single_thread()

        self.elapsed = time.time() - a

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        if self.pool is not None:
            self.pool.terminate()
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled!')
        self.done_signal.emit()



# -*- coding: utf-8 -*-
# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import time
import scipy.sparse as sp
import numpy as np

from GridCal.Engine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults

linear_solver = get_linear_solver()
sparse = get_sparse_type()


def dcpf(Ybus, Bpqpv, Bref, Sbus, Ibus, V0, ref, pvpq, pq, pv) -> NumericPowerFlowResults:
    """
    Solves a DC power flow.
    :param Ybus: Normal circuit admittance matrix
    :param Sbus: Complex power injections at all the nodes
    :param Ibus: Complex current injections at all the nodes
    :param V0: Array of complex seed voltage (it contains the ref voltages)
    :param ref: array of the indices of the slack nodes
    :param pvpq: array of the indices of the non-slack nodes
    :param pq: array of the indices of the pq nodes
    :param pv: array of the indices of the pv nodes
    :return:
        Complex voltage solution
        Converged: Always true
        Solution error
        Computed power injections given the found solution
    """

    start = time.time()
    npq = len(pq)
    npv = len(pv)
    if (npq + npv) > 0:
        # Decompose the voltage in angle and magnitude
        Va_ref = np.angle(V0[ref])  # we only need the angles at the slack nodes
        Vm = np.abs(V0)

        # initialize result vector
        Va = np.empty(len(V0))

        # compose the reduced power injections
        # Since we have removed the slack nodes, we must account their influence as injections Bref * Va_ref
        Pinj = Sbus[pvpq].real + (- Bref * Va_ref + Ibus[pvpq].real) * Vm[pvpq]

        # update angles for non-reference buses
        Va[pvpq] = linear_solver(Bpqpv, Pinj)
        Va[ref] = Va_ref

        # re assemble the voltage
        V = Vm * np.exp(1j * Va)

        # compute the calculated power injection and the error of the voltage solution
        Scalc = V * np.conj(Ybus * V - Ibus)

        # compute the power mismatch between the specified power Sbus and the calculated power Scalc
        mis = Scalc - Sbus  # complex power mismatch
        mismatch = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

        # check for convergence
        norm_f = np.linalg.norm(mismatch, np.Inf)
    else:
        norm_f = 0.0
        V = V0
        Scalc = V * np.conj(Ybus * V - Ibus)

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, True, norm_f, Scalc, None, None, None, 1, elapsed)


def lacpf(Y, Ys, S, I, Vset, pq, pv) -> NumericPowerFlowResults:
    """
    Linearized AC Load Flow

    form the article:

    Linearized AC Load Flow Applied to Analysis in Electric Power Systems
        by: P. Rossoni, W. M da Rosa and E. A. Belati
    Args:
        Y: Admittance matrix
        Ys: Admittance matrix of the series elements
        S: Power injections vector of all the nodes
        Vset: Set voltages of all the nodes (used for the slack and PV nodes)
        pq: list of indices of the pq nodes
        pv: list of indices of the pv nodes

    Returns: Voltage vector, converged?, error, calculated power and elapsed time
    """

    start = time.time()

    pvpq = np.r_[pv, pq]
    npq = len(pq)
    npv = len(pv)

    if (npq + npv) > 0:
        # compose the system matrix
        # G = Y.real
        # B = Y.imag
        # Gp = Ys.real
        # Bp = Ys.imag

        A11 = -Ys.imag[np.ix_(pvpq, pvpq)]
        A12 = Y.real[np.ix_(pvpq, pq)]
        A21 = -Ys.real[np.ix_(pq, pvpq)]
        A22 = -Y.imag[np.ix_(pq, pq)]

        Asys = sp.vstack([sp.hstack([A11, A12]),
                          sp.hstack([A21, A22])], format="csc")

        # compose the right hand side (power vectors)
        rhs = np.r_[S.real[pvpq], S.imag[pq]]

        # solve the linear system
        try:
            x = linear_solver(Asys, rhs)
        except Exception as e:
            voltages_vector = Vset
            # Calculate the error and check the convergence
            s_calc = voltages_vector * np.conj(Y * voltages_vector)
            # complex power mismatch
            power_mismatch = s_calc - S
            # concatenate error by type
            mismatch = np.r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]
            # check for convergence
            norm_f = np.linalg.norm(mismatch, np.Inf)
            end = time.time()
            elapsed = end - start
            return voltages_vector, False, norm_f, s_calc, 1, elapsed

        # compose the results vector
        voltages_vector = Vset.copy()

        #  set the pv voltages
        va_pv = x[0:npv]
        vm_pv = np.abs(Vset[pv])
        voltages_vector[pv] = vm_pv * np.exp(1.0j * va_pv)

        # set the PQ voltages
        va_pq = x[npv:npv+npq]
        vm_pq = np.ones(npq) - x[npv+npq::]
        voltages_vector[pq] = vm_pq * np.exp(1.0j * va_pq)

        # Calculate the error and check the convergence
        s_calc = voltages_vector * np.conj(Y * voltages_vector)
        # complex power mismatch
        power_mismatch = s_calc - S
        # concatenate error by type
        mismatch = np.r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

        # check for convergence
        norm_f = np.linalg.norm(mismatch, np.Inf)
    else:
        norm_f = 0.0
        voltages_vector = Vset
        s_calc = voltages_vector * np.conj(Y * voltages_vector)

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(voltages_vector, True, norm_f, s_calc, None, None, None, 1, elapsed)






class Branch:
    """
    Branch class template
    """

    def __init__(self, name, bus_from, bus_to, rating):
        self.name = name

        self.f = bus_from

        self.t = bus_to

        self.rating = rating

    def get_ABCD(self, Sbase):
        pass


# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import pandas as pd
import numpy as np
import GridCal.Engine.basic_structures as bs

import GridCal.Engine.Simulations.PowerFlow.linearized_power_flow as aclin
import GridCal.Engine.Simulations.PowerFlow.jacobian_based_power_flow as acjb
import GridCal.Engine.Simulations.PowerFlow.jacobian_based_acdc_power_flow as acdcjb
import GridCal.Engine.Simulations.PowerFlow.fast_decoupled_power_flow as acfd
import GridCal.Engine.Simulations.PowerFlow.helm_power_flow as hl

from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
from GridCal.Engine.Core.snapshot_pf_data import SnapshotData
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit


class ConvergenceReport:

    def __init__(self):
        self.methods_ = list()
        self.converged_ = list()
        self.error_ = list()
        self.elapsed_ = list()
        self.iterations_ = list()

    def add(self, method, converged, error, elapsed, iterations):
        self.methods_.append(method)
        self.converged_.append(converged)
        self.error_.append(error)
        self.elapsed_.append(elapsed)
        self.iterations_.append(iterations)

    def converged(self):
        if len(self.converged_) > 0:
            return self.converged_[-1]
        else:
            return False

    def error(self):
        if len(self.error_) > 0:
            return self.error_[-1]
        else:
            return 0

    def elapsed(self):
        if len(self.elapsed_) > 0:
            return self.elapsed_[-1]
        else:
            return 0

    def to_dataframe(self):
        data = {'Method': self.methods_,
                'Converged?': self.converged_,
                'Error': self.error_,
                'Elapsed (s)': self.elapsed_,
                'Iterations': self.iterations_}

        df = pd.DataFrame(data)

        return df


def solve(circuit: SnapshotData, options: PowerFlowOptions, report: ConvergenceReport, V0, Sbus, Ibus,
          pq, pv, ref, pqpv, logger=bs.Logger()) -> NumericPowerFlowResults:
    """
    Run a power flow simulation using the selected method (no outer loop controls).
    :param circuit: SnapshotData circuit, this ensures on-demand admittances computation
    :param options: PowerFlow options
    :param report: Convergence report to fill in
    :param V0: Array of initial voltages
    :param Sbus: Array of power injections
    :param Ibus: Array of current injections
    :param pq: Array of pq nodes
    :param pv: Array of pv nodes
    :param ref: Array of slack nodes
    :param pqpv: Array of (sorted) pq and pv nodes
    :param logger: Logger
    :return: NumericPowerFlowResults 
    """

    if options.retry_with_other_methods:
        if circuit.any_control:
            solver_list = [bs.SolverType.NR,
                           bs.SolverType.LM,
                           bs.SolverType.HELM,
                           bs.SolverType.IWAMOTO,
                           bs.SolverType.LACPF]
        else:
            solver_list = [bs.SolverType.NR,
                           bs.SolverType.HELM,
                           bs.SolverType.IWAMOTO,
                           bs.SolverType.LM,
                           bs.SolverType.LACPF]

        if options.solver_type in solver_list:
            solver_list.remove(options.solver_type)

        solvers = [options.solver_type] + solver_list
    else:
        # No retry selected
        solvers = [options.solver_type]

    # set worked to false to enter in the loop
    solver_idx = 0

    # set the initial value

    final_solution = NumericPowerFlowResults(V=V0,
                                             converged=False,
                                             norm_f=1e200,
                                             Scalc=Sbus,
                                             ma=circuit.branch_data.m[:, 0],
                                             theta=circuit.branch_data.theta[:, 0],
                                             Beq=circuit.branch_data.Beq[:, 0],
                                             iterations=0,
                                             elapsed=0)

    while solver_idx < len(solvers) and not final_solution.converged:
        # get the solver
        solver_type = solvers[solver_idx]

        # type HELM
        if solver_type == bs.SolverType.HELM:
            solution = hl.helm_josep(Ybus=circuit.Ybus,
                                     Yseries=circuit.Yseries,
                                     V0=V0,  # take V0 instead of V
                                     S0=Sbus,
                                     Ysh0=circuit.Yshunt,
                                     pq=pq,
                                     pv=pv,
                                     sl=ref,
                                     pqpv=pqpv,
                                     tolerance=options.tolerance,
                                     max_coeff=options.max_iter,
                                     use_pade=True,
                                     verbose=False)

        # type DC
        elif solver_type == bs.SolverType.DC:
            solution = aclin.dcpf(Ybus=circuit.Ybus,
                                  Bpqpv=circuit.Bpqpv,
                                  Bref=circuit.Bref,
                                  Sbus=Sbus,
                                  Ibus=Ibus,
                                  V0=V0,
                                  ref=ref,
                                  pvpq=pqpv,
                                  pq=pq,
                                  pv=pv)

        # LAC PF
        elif solver_type == bs.SolverType.LACPF:
            solution = aclin.lacpf(Y=circuit.Ybus,
                                   Ys=circuit.Yseries,
                                   S=Sbus,
                                   I=Ibus,
                                   Vset=V0,
                                   pq=pq,
                                   pv=pv)

        # Levenberg-Marquardt
        elif solver_type == bs.SolverType.LM:
            if circuit.any_control:
                solution = acdcjb.LM_ACDC(nc=circuit,
                                          Vbus=V0,
                                          Sbus=Sbus,
                                          tolerance=options.tolerance,
                                          max_iter=options.max_iter)
            else:
                solution = acjb.levenberg_marquardt_pf(Ybus=circuit.Ybus,
                                                       Sbus_=Sbus,
                                                       V0=final_solution.V,
                                                       Ibus=Ibus,
                                                       pv_=pv,
                                                       pq_=pq,
                                                       Qmin=circuit.Qmin_bus[0, :],
                                                       Qmax=circuit.Qmax_bus[0, :],
                                                       tol=options.tolerance,
                                                       max_it=options.max_iter,
                                                       control_q=options.control_Q)

        # Fast decoupled
        elif solver_type == bs.SolverType.FASTDECOUPLED:
            solution = acfd.FDPF(Vbus=V0,
                                 Sbus=Sbus,
                                 Ibus=Ibus,
                                 Ybus=circuit.Ybus,
                                 B1=circuit.B1,
                                 B2=circuit.B2,
                                 pq=pq,
                                 pv=pv,
                                 pqpv=pqpv,
                                 tol=options.tolerance,
                                 max_it=options.max_iter)

        # Newton-Raphson (full)
        elif solver_type == bs.SolverType.NR:

            if circuit.any_control:
                # Solve NR with the AC/DC algorithm
                solution = acdcjb.NR_LS_ACDC(nc=circuit,
                                             Vbus=V0,
                                             Sbus=Sbus,
                                             tolerance=options.tolerance,
                                             max_iter=options.max_iter,
                                             acceleration_parameter=options.backtracking_parameter,
                                             mu_0=options.mu,
                                             control_q=options.control_Q)
            else:
                # Solve NR with the AC algorithm
                solution = acjb.NR_LS(Ybus=circuit.Ybus,
                                      Sbus_=Sbus,
                                      V0=final_solution.V,
                                      Ibus=Ibus,
                                      pv_=pv,
                                      pq_=pq,
                                      Qmin=circuit.Qmin_bus[0, :],
                                      Qmax=circuit.Qmax_bus[0, :],
                                      tol=options.tolerance,
                                      max_it=options.max_iter,
                                      mu_0=options.mu,
                                      acceleration_parameter=options.backtracking_parameter,
                                      control_q=options.control_Q)

        # Newton-Raphson-Decpupled
        elif solver_type == bs.SolverType.NRD:
            # Solve NR with the linear AC solution
            solution = acjb.NRD_LS(Ybus=circuit.Ybus,
                                   Sbus=Sbus,
                                   V0=final_solution.V,
                                   Ibus=Ibus,
                                   pv=pv,
                                   pq=pq,
                                   tol=options.tolerance,
                                   max_it=options.max_iter,
                                   acceleration_parameter=options.backtracking_parameter)

        # Newton-Raphson-Iwamoto
        elif solver_type == bs.SolverType.IWAMOTO:
            solution = acjb.IwamotoNR(Ybus=circuit.Ybus,
                                      Sbus_=Sbus,
                                      V0=final_solution.V,
                                      Ibus=Ibus,
                                      pv_=pv,
                                      pq_=pq,
                                      Qmin=circuit.Qmin_bus[0, :],
                                      Qmax=circuit.Qmax_bus[0, :],
                                      tol=options.tolerance,
                                      max_it=options.max_iter,
                                      control_q=options.control_Q,
                                      robust=True)

        # Newton-Raphson in current equations
        elif solver_type == bs.SolverType.NRI:
            solution = acjb.NR_I_LS(Ybus=circuit.Ybus,
                                    Sbus_sp=Sbus,
                                    V0=final_solution.V,
                                    Ibus_sp=Ibus,
                                    pv=pv,
                                    pq=pq,
                                    tol=options.tolerance,
                                    max_it=options.max_iter)

        else:
            # for any other method, raise exception
            raise Exception(solver_type + ' Not supported in power flow mode')

        # record the method used, if it improved the solution
        if solution.norm_f < final_solution.norm_f:
            report.add(method=solver_type,
                       converged=solution.converged,
                       error=solution.norm_f,
                       elapsed=solution.elapsed,
                       iterations=solution.iterations)
            final_solution = solution

        # record the solver steps
        solver_idx += 1

    if not final_solution.converged:
        logger.add_error('Did not converge, even after retry!', 'Error', str(final_solution.norm_f), options.tolerance)

    if final_solution.ma is None:
        final_solution.ma = circuit.branch_data.m[:, 0]

    if final_solution.theta is None:
        final_solution.theta = circuit.branch_data.theta[:, 0]

    if final_solution.Beq is None:
        final_solution.Beq = circuit.branch_data.Beq[:, 0]

    return final_solution


def outer_loop_power_flow(circuit: SnapshotData, options: PowerFlowOptions,
                          voltage_solution, Sbus, Ibus, branch_rates, t=0,
                          logger=bs.Logger()) -> "PowerFlowResults":
    """
    Run a power flow simulation for a single circuit using the selected outer loop
    controls. This method shouldn't be called directly.
    :param circuit: CalculationInputs instance
    :param options:
    :param voltage_solution: vector of initial voltages
    :param Sbus: vector of power injections
    :param Ibus: vector of current injections
    :param branch_rates:
    :param t: time step
    :param logger:
    :return: PowerFlowResults instance
    """

    # get the original types and compile this class' own lists of node types for thread independence
    bus_types = circuit.bus_types.copy()

    # vd = circuit.vd.copy()
    # pq = circuit.pq.copy()
    # pv = circuit.pv.copy()
    # pqpv = circuit.pqpv.copy()

    report = ConvergenceReport()
    solution = NumericPowerFlowResults(V=voltage_solution,
                                       converged=False,
                                       norm_f=1e200,
                                       Scalc=Sbus,
                                       ma=circuit.branch_data.m[:, 0],
                                       theta=circuit.branch_data.theta[:, 0],
                                       Beq=circuit.branch_data.Beq[:, 0],
                                       iterations=0,
                                       elapsed=0)

    # this the "outer-loop"
    if len(circuit.vd) == 0:
        voltage_solution = np.zeros(len(Sbus), dtype=complex)
        normF = 0
        Scalc = Sbus.copy()
        any_q_control_issue = False
        converged = True
        logger.add_error('Not solving power flow because there is no slack bus')
    else:

        # run the power flow method that shall be run
        solution = solve(circuit=circuit,
                         options=options,
                         report=report,  # is modified here
                         V0=voltage_solution,
                         Sbus=Sbus,
                         Ibus=Ibus,
                         pq=circuit.pq,
                         pv=circuit.pv,
                         ref=circuit.vd,
                         pqpv=circuit.pqpv,
                         logger=logger)

        if options.distributed_slack:
            # Distribute the slack power
            slack_power = Sbus[circuit.vd].real.sum()
            total_installed_power = circuit.bus_installed_power.sum()

            if total_installed_power > 0.0:
                delta = slack_power * circuit.bus_installed_power / total_installed_power

                # repeat power flow with the redistributed power
                solution = solve(circuit=circuit,
                                 options=options,
                                 report=report,  # is modified here
                                 V0=solution.V,
                                 Sbus=Sbus + delta,
                                 Ibus=Ibus,
                                 pq=circuit.pq,
                                 pv=circuit.pv,
                                 ref=circuit.vd,
                                 pqpv=circuit.pqpv,
                                 logger=logger)

    # Compute the branches power and the slack buses power
    Sfb, Stb, If, It, Vbranch, loading, losses, \
     flow_direction, Sbus = power_flow_post_process(calculation_inputs=circuit,
                                                    Sbus=solution.Scalc,
                                                    V=solution.V,
                                                    branch_rates=branch_rates)

    # voltage, Sf, loading, losses, error, converged, Qpv
    results = PowerFlowResults(n=circuit.nbus,
                               m=circuit.nbr,
                               n_tr=circuit.ntr,
                               n_hvdc=circuit.nhvdc,
                               bus_names=circuit.bus_names,
                               branch_names=circuit.branch_names,
                               transformer_names=circuit.tr_names,
                               hvdc_names=circuit.hvdc_names,
                               bus_types=bus_types)

    results.Sbus = solution.Scalc * circuit.Sbase  # MVA
    results.voltage = solution.V
    results.Sf = Sfb  # in MVA already
    results.St = Stb  # in MVA already
    results.If = If  # in p.u.
    results.It = It  # in p.u.
    results.ma = solution.ma
    results.theta = solution.theta
    results.Beq = solution.Beq
    results.Vbranch = Vbranch
    results.loading = loading
    results.losses = losses
    results.flow_direction = flow_direction
    results.transformer_tap_module = solution.ma[circuit.transformer_idx]
    results.convergence_reports.append(report)
    results.Qpv = Sbus.imag[circuit.pv]

    # HVDC results are gathered in the multi island power flow function due to their nature

    return results


def power_flow_post_process(calculation_inputs: SnapshotData, Sbus, V, branch_rates):
    """
    Compute the power flows trough the branches.

    Arguments:

        **calculation_inputs**: instance of Circuit

        **V**: Voltage solution array for the circuit buses

        **only_power**: compute only the power injection

    Returns:

        Sf (MVA), If (p.u.), loading (p.u.), losses (MVA), Sbus(MVA)
    """
    # Compute the slack and pv buses power
    vd = calculation_inputs.vd
    pv = calculation_inputs.pv

    # power at the slack nodes
    Sbus[vd] = V[vd] * np.conj(calculation_inputs.Ybus[vd, :].dot(V))

    # Reactive power at the pv nodes
    P = Sbus[pv].real
    Q = (V[pv] * np.conj(calculation_inputs.Ybus[pv, :].dot(V))).imag
    Sbus[pv] = P + 1j * Q  # keep the original P injection and set the calculated reactive power

    # Branches current, loading, etc
    Vf = calculation_inputs.Cf * V
    Vt = calculation_inputs.Ct * V
    If = calculation_inputs.Yf * V
    It = calculation_inputs.Yt * V
    Sf = Vf * np.conj(If)
    St = Vt * np.conj(It)

    # Branch losses in MVA
    losses = (Sf + St) * calculation_inputs.Sbase

    flow_direction = Sf.real / np.abs(Sf + 1e-20)

    # branch voltage increment
    Vbranch = Vf - Vt

    # Branch power in MVA
    Sfb = Sf * calculation_inputs.Sbase
    Stb = St * calculation_inputs.Sbase

    # Branch loading in p.u.
    loading = Sfb / (branch_rates + 1e-9)

    return Sfb, Stb, If, It, Vbranch, loading, losses, flow_direction, Sbus


def single_island_pf(circuit: SnapshotData, Vbus, Sbus, Ibus, branch_rates,
                     options: PowerFlowOptions, logger: bs.Logger) -> "PowerFlowResults":
    """
    Run a power flow for a circuit. In most cases, the **run** method should be used instead.
    :param circuit: SnapshotData instance
    :param Vbus: Initial voltage at each bus in complex per unit
    :param Sbus: Power injection at each bus in complex MVA
    :param Ibus: Current injection at each bus in complex MVA
    :param branch_rates: array of branch rates
    :param options: PowerFlowOptions instance
    :param logger: Logger instance
    :return: PowerFlowResults instance
    """

    # solve the power flow
    results = outer_loop_power_flow(circuit=circuit,
                                    options=options,
                                    voltage_solution=Vbus,
                                    Sbus=Sbus,
                                    Ibus=Ibus,
                                    branch_rates=branch_rates,
                                    logger=logger)

    # did it worked?
    worked = np.all(results.converged)

    # if not worked:
    #     logger.add_error('Did not converge, even after retry!', 'Error', str(results.error), options.tolerance)

    return results


def multi_island_pf(multi_circuit: MultiCircuit, options: PowerFlowOptions, opf_results=None,
                    logger=bs.Logger()) -> "PowerFlowResults":
    """
    Multiple islands power flow (this is the most generic power flow function)
    :param multi_circuit: MultiCircuit instance
    :param options: PowerFlowOptions instance
    :param opf_results: OPF results, to be used if not None
    :param logger: list of events to add to
    :return: PowerFlowResults instance
    """

    nc = compile_snapshot_circuit(circuit=multi_circuit,
                                  apply_temperature=options.apply_temperature_correction,
                                  branch_tolerance_mode=options.branch_impedance_tolerance_mode,
                                  opf_results=opf_results)

    calculation_inputs = nc.split_into_islands(ignore_single_node_islands=options.ignore_single_node_islands)

    results = PowerFlowResults(n=nc.nbus,
                               m=nc.nbr,
                               n_tr=nc.ntr,
                               n_hvdc=nc.nhvdc,
                               bus_names=nc.bus_data.bus_names,
                               branch_names=nc.branch_data.branch_names,
                               transformer_names=nc.transformer_data.tr_names,
                               hvdc_names=nc.hvdc_data.names,
                               bus_types=nc.bus_data.bus_types)

    if len(calculation_inputs) > 1:

        # simulate each island and merge the results
        for i, calculation_input in enumerate(calculation_inputs):

            if len(calculation_input.vd) > 0:

                # run circuit power flow
                res = single_island_pf(circuit=calculation_input,
                                       Vbus=calculation_input.Vbus,
                                       Sbus=calculation_input.Sbus,
                                       Ibus=calculation_input.Ibus,
                                       branch_rates=calculation_input.Rates,
                                       options=options,
                                       logger=logger)

                bus_original_idx = calculation_input.original_bus_idx
                branch_original_idx = calculation_input.original_branch_idx
                tr_original_idx = calculation_input.original_tr_idx

                # merge the results from this island
                results.apply_from_island(res, bus_original_idx, branch_original_idx, tr_original_idx)

            else:
                logger.add_info('No slack nodes in the island', str(i))
    else:

        if len(calculation_inputs[0].vd) > 0:
            # only one island
            # run circuit power flow
            res = single_island_pf(circuit=calculation_inputs[0],
                                   Vbus=calculation_inputs[0].Vbus,
                                   Sbus=calculation_inputs[0].Sbus,
                                   Ibus=calculation_inputs[0].Ibus,
                                   branch_rates=calculation_inputs[0].Rates,
                                   options=options,
                                   logger=logger)

            if calculation_inputs[0].nbus == nc.nbus:
                # we can confidently say that the island is the only one
                results = res
            else:
                # the island is the only valid subset, but does not contain all the buses
                bus_original_idx = calculation_inputs[0].original_bus_idx
                branch_original_idx = calculation_inputs[0].original_branch_idx
                tr_original_idx = calculation_inputs[0].original_tr_idx

                # merge the results from this island
                results.apply_from_island(res, bus_original_idx, branch_original_idx, tr_original_idx)

        else:
            logger.add_error('There are no slack nodes')

    # compile HVDC results (available for the complete grid since HVDC line as formulated are split objects
    # Pt is the "generation" at the sending point
    results.hvdc_Pf = -nc.hvdc_Pf
    results.hvdc_Pt = -nc.hvdc_Pt
    results.hvdc_loading = nc.hvdc_loading
    results.hvdc_losses = nc.hvdc_losses

    return results


def power_flow_worker_args(args):
    """
    Power flow worker to schedule parallel power flows

    args -> t, options: PowerFlowOptions, circuit: Circuit, Vbus, Sbus, Ibus, return_dict


        **t: execution index
        **options: power flow options
        **circuit: circuit
        **Vbus: Voltages to initialize
        **Sbus: Power injections
        **Ibus: Current injections
        **return_dict: parallel module dictionary in which to return the values
    :return:
    """
    t, options, circuit, Vbus, Sbus, Ibus, branch_rates = args

    res = single_island_pf(circuit=circuit,
                           Vbus=Vbus,
                           Sbus=Sbus,
                           Ibus=Ibus,
                           branch_rates=branch_rates,
                           options=options,
                           logger=bs.Logger())

    return t, res



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import json
import numpy as np
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_model import ResultsModel


class NMinusKResults:

    def __init__(self, n, m, bus_names, branch_names, bus_types):
        """
        TimeSeriesResults constructor
        @param n: number of buses
        @param m: number of branches
        """

        self.name = 'N-1'

        self.bus_types = np.zeros(n, dtype=int)

        self.branch_names = branch_names

        self.bus_names = bus_names

        self.bus_types = bus_types

        self.voltage = np.ones((m, n), dtype=complex)

        self.S = np.zeros((m, n), dtype=complex)

        self.Sf = np.zeros((m, m), dtype=complex)

        self.loading = np.zeros((m, m), dtype=complex)

        self.otdf = np.zeros((m, m))

        self.available_results = [ResultTypes.OTDF,
                                  ResultTypes.BusActivePower,
                                  ResultTypes.BranchActivePowerFrom,
                                  ResultTypes.BranchLoading]

    def get_steps(self):
        return

    def get_results_dict(self):
        """
        Returns a dictionary with the results sorted in a dictionary
        :return: dictionary of 2D numpy arrays (probably of complex numbers)
        """
        data = {'Vm': np.abs(self.voltage).tolist(),
                'Va': np.angle(self.voltage).tolist(),
                'P': self.S.real.tolist(),
                'Q': self.S.imag.tolist(),
                'Sbr_real': self.Sf.real.tolist(),
                'Sbr_imag': self.Sf.imag.tolist(),
                'loading': np.abs(self.loading).tolist()}
        return data

    def save(self, fname):
        """
        Export as json
        """
        with open(fname, "w") as output_file:
            json_str = json.dumps(self.get_results_dict())
            output_file.write(json_str)

    def mdl(self, result_type: ResultTypes):
        """
        Plot the results
        :param result_type:
        :return:
        """

        index = self.branch_names

        if result_type == ResultTypes.BusVoltageModule:
            data = np.abs(self.voltage)
            y_label = '(p.u.)'
            title = 'Bus voltage '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BusVoltageAngle:
            data = np.angle(self.voltage, deg=True)
            y_label = '(Deg)'
            title = 'Bus voltage '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BusActivePower:
            data = self.S.real
            y_label = '(MW)'
            title = 'Bus active power '
            labels = self.bus_names
            # index = self.branch_names

        elif result_type == ResultTypes.BranchActivePowerFrom:
            data = self.Sf.real
            y_label = 'MW'
            title = 'Branch active power '
            labels = ['# ' + x for x in self.branch_names]
            # index = self.branch_names

        elif result_type == ResultTypes.BranchLoading:
            data = self.loading.real * 100
            y_label = '(%)'
            title = 'Branch loading '
            labels = ['# ' + x for x in self.branch_names]
            # index = self.branch_names

        elif result_type == ResultTypes.OTDF:
            data = self.otdf
            y_label = 'Per unit'
            labels = ['# ' + x for x in self.branch_names]

            title = 'OTDF'
        else:
            raise Exception('Result type not understood:' + str(result_type))

        # assemble model
        mdl = ResultsModel(data=data,
                           index=index,
                           columns=labels,
                           title=title,
                           ylabel=y_label)
        return mdl




import numpy as np

from research.three_phase.Engine.general import *


class Generator:

    def __init__(self, name, P, v=1.0, conn=Phases.ABC, conn_type=Connection.Delta):
        """
        Generator constructor
        :param name: name of the element
        :param P: active power array in MW
        :param v: voltage set point (single value)
        :param conn: phases of connection
        :param conn_type: type of connection (Wye or Delta)
        """

        self.name = name

        # active power in MW
        self.P = P

        # voltage set point in per unit
        self.v = v

        # connection type
        self.conn_type = conn_type

        # vector of the used phases
        self.phases = np.array(conn)

        # number of phases
        self.number_of_phases = len(self.phases)

        if self.conn_type != Connection.PositiveSequence:
            if len(self.P) != 3:
                raise Exception(self.name + ': The power array must have 3 values '
                                            'since the connection implied a 3-phase system')
            else:
                pass
        else:
            pass

    def get_values(self):
        """
        Returns the wye-connected power and the voltage set point
        :return: Wye-connected power
        """
        if self.number_of_phases == 3:
            if self.conn_type == Connection.Wye:
                return self.P, self.v
            else:
                return delta_to_wye(self.P), self.v
        else:
            return self.P, self.v




from matplotlib import pyplot as plt
import numpy as np
plt.ion()



def replace_non_ascii(text):
    return ''.join([i if ord(i) < 128 else ' ' for i in text])


def plot_full_convergence(err, powerInjections, S, title, ext='.eps', save=True):
    plt.figure()

    titles = ['Maximum mismatch', 'PV nodes P mismatch', 'PV nodes Q mismatch', 'PQ nodes P mismatch', 'PQ nodes Q mismatch']
    idx = [0, 1, 3, 4]
    k = 0
    for i in idx:
        k += 1
        plt.subplot(2, 2, k)
        if not np.all(err[i] == 0):
            plt.plot(err[i])
            plt.title(titles[i])
            try:
                plt.yscale('log')
            except:
                print()

    fig = plt.gcf()
    fig.suptitle(title, fontsize=14)

    if save:
        plt.savefig(replace_non_ascii(title).replace(" ", "_") + ext)
    else:
        plt.show()


def plot_errors(err1, err2, title1, title2, title, ext = '.eps', save=True):
    plt.figure()
    plt.plot(err1[0], marker='x', label=title1)
    plt.plot(err2[0], marker='x', label=title2)
    plt.yscale('log')
    plt.ylabel('Error')
    plt.xlabel('Iterations')
    plt.title(title)
    plt.legend()
    if save:
        plt.savefig(replace_non_ascii(title).replace(" ", "_") + ext)
    else:
        plt.show()


def plot_error(err1, title1, title, ext = '.eps', save=True):
    plt.figure()
    plt.plot(err1[0], marker='x', label=title1)
    plt.yscale('log')
    plt.ylabel('Error')
    plt.xlabel('Iterations')
    plt.title(title)
    plt.legend()
    if save:
        plt.savefig(replace_non_ascii(title).replace(" ", "_") + ext)
    else:
        plt.show()


def plot_voltages(Vlst, title, C=None, ext = '.eps', save=True):

    if C is None:
        a = 1
        b = 2
    else:
        a = 2
        b = 2
    V = np.array(Vlst)
    Vm = np.abs(V)
    Va = np.angle(V, deg=True)
    plt.figure(figsize=(9, 5))
    plt.subplot(a, b, 1)
    plt.plot(Vm, label='Module')
    plt.ylabel('Voltage module')
    plt.xlabel('Iterations')

    plt.subplot(a, b, 2)
    plt.plot(Va, label='Angle in degrees')
    plt.ylabel('Voltage angle (degrees)')
    plt.xlabel('Iterations')

    if C is not None:
        n = len(C)
        plt.subplot(a, b, 3)
        # print(n)
        for i in range(n):
            # print(i)
            # print((C[i, :]))
            plt.plot(np.abs(C[i, :]))
        plt.ylabel('Voltage coefficients ')
        plt.xlabel('Iterations')

        plt.subplot(a, b, 4)
        for i in range(n):
            plt.plot(np.angle(C[i, :], deg=True))
        plt.ylabel('Voltage coefficients angle (degrees)')
        plt.xlabel('Iterations')

    fig = plt.gcf()
    fig.suptitle(title)
    if save:
        plt.savefig(replace_non_ascii(title).replace(" ", "_") + ext)
    else:
        plt.show()


if __name__ == "__main__":
    # global vars
    use_sparse = True
    cmax = 40
    usepade = True
    ext = '.eps'

    ########################################################################################################################
    # Pade VS Epsilon
    ########################################################################################################################
    print('Grid load')
    admittances, powerInjections, bustype, \
    voltageSetPoints, slackIndices, nbus, eps = grids.Lynn_example2(use_sparse)
    gridname = '5-node grid'

    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err_pade, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=True)

    V1, C, W, X, R, H, Yred, err_eps, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=False)

    plot_errors(err_pade, err_eps, 'Padè', "Wynn's epsilon", 'Padè VS Epsilon algorithm', ext)

    ########################################################################################################################
    # Lynn 2
    ########################################################################################################################

    print('Grid load')
    admittances, powerInjections, bustype, \
    voltageSetPoints, slackIndices, nbus, eps = grids.Lynn_example2(use_sparse)
    gridname = '5-node grid'

    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=usepade)

    plot_full_convergence(err, powerInjections, S, 'HELM-Z full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)

    # Full convergence characteristics on Lynn 2 (HELM-ASU)
    print('HELM ASU FUll convergence')
    V, Ymat, F, C, W, Q, err, converged, S, Vlst = helm_asu(admittances, slackIndices, cmax, powerInjections,
                                                            voltageSetPoints, bustype, eps, usePade=usepade, useFFT=False)

    try:
        plot_full_convergence(err, powerInjections, S, 'HELM-ASU full convergence characteristic ('+gridname+')', ext)
        plot_voltages(Vlst, 'HELM-ASU voltage convergence ('+gridname+')', ext=ext)
    except:
        print()
    ########################################################################################################################
    # IEEE 5
    ########################################################################################################################

    print('Grid load')
    admittances, B, powerInjections, bustype, voltageSetPoints, slackIndices, nbus, eps = grids.IEEE_5(use_sparse)
    gridname = 'IEEE5'

    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=usepade)

    plot_full_convergence(err, powerInjections, S, 'HELM-Z full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)

    # Full convergence characteristics on Lynn 2 (HELM-ASU)
    print('HELM ASU FUll convergence')
    V, Ymat, F, C, W, Q, err, converged, S, Vlst = helm_asu(admittances, slackIndices, cmax, powerInjections,
                                                            voltageSetPoints, bustype, eps, usePade=usepade, useFFT=False)

    plot_full_convergence(err, powerInjections, S, 'HELM-ASU full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-ASU voltage convergence ('+gridname+')', ext=ext)

    ########################################################################################################################
    # IEEE 14
    ########################################################################################################################
    admittances, B, powerInjections, bustype, voltageSetPoints, slackIndices, nbus, eps = grids.IEEE_14(use_sparse)
    gridname = 'IEEE14'
    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=usepade)

    plot_full_convergence(err, powerInjections, S, 'HELM-Z full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)

    # Full convergence characteristics on Lynn 2 (HELM-ASU)
    print('HELM ASU FUll convergence')
    V, Ymat, F, C, W, Q, err, converged, S, Vlst = helm_asu(admittances, slackIndices, cmax, powerInjections,
                                                            voltageSetPoints, bustype, eps, usePade=usepade, useFFT=False)

    plot_full_convergence(err, powerInjections, S, 'HELM-ASU full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-ASU voltage convergence ('+gridname+')', ext=ext)

    ########################################################################################################################
    # IEEE 30
    ########################################################################################################################
    admittances, B, powerInjections, bustype, voltageSetPoints, slackIndices, nbus, eps = grids.IEEE_30(use_sparse)
    gridname = 'IEEE30'
    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=usepade)

    plot_full_convergence(err, powerInjections, S, 'HELM-Z full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)

    # Full convergence characteristics on Lynn 2 (HELM-ASU)
    print('HELM ASU FUll convergence')
    V, Ymat, F, C, W, Q, err, converged, S, Vlst = helm_asu(admittances, slackIndices, cmax, powerInjections,
                                                            voltageSetPoints, bustype, eps, usePade=usepade, useFFT=False)

    plot_full_convergence(err, powerInjections, S, 'HELM-ASU full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-ASU voltage convergence ('+gridname+')', ext=ext)

    ########################################################################################################################
    # IEEE 118
    ########################################################################################################################
    admittances, B, powerInjections, bustype, voltageSetPoints, slackIndices, nbus, eps = grids.IEEE_118(use_sparse)
    gridname = 'IEEE118'
    # Full convergence characteristics on Lynn 2 (HELM-Z)
    print('HELM Z FUll convergence')
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=usepade)

    plot_full_convergence(err, powerInjections, S, 'HELM-Z full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)

    # Full convergence characteristics on Lynn 2 (HELM-ASU)
    print('HELM ASU FUll convergence')
    V, Ymat, F, C, W, Q, err, converged, S, Vlst = helm_asu(admittances, slackIndices, cmax, powerInjections,
                                                            voltageSetPoints, bustype, eps, usePade=usepade, useFFT=False)

    plot_full_convergence(err, powerInjections, S, 'HELM-ASU full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-ASU voltage convergence ('+gridname+')', ext=ext)

    ########################################################################################################################
    # IEEE 118 (PV pre approximation)
    ########################################################################################################################
    admittances, B, powerInjections, bustype, voltageSetPoints, slackIndices, nbus, eps = grids.IEEE_30(use_sparse)
    gridname = 'IEEE30'

    cmax_red = 15
    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices, cmax_red, powerInjections,
                               voltageSetPoints, bustype, eps, usePade=True)

    pv_idx = np.where(bustype == 2)[0]
    Va = np.angle(V1)
    # Va[pv_idx] *= -1
    Vm = np.abs(voltageSetPoints)
    V0 = Vm * np.exp(1j * Va)

    # set the PV to VD
    bustype_ = bustype.copy()
    bustype_[pv_idx] = 3
    slackIndices_ = np.r_[slackIndices, pv_idx]
    slackIndices_.sort()


    V1, C, W, X, R, H, Yred, err, converged_, \
    best_err, S, Vlst = helm_z(admittances, slackIndices_, cmax, powerInjections,
                               V0, bustype_, eps, usePade=True, inherited_pv=pv_idx)


    plot_full_convergence(err, powerInjections, S, 'HELM-Z (PV to Slack) full convergence characteristic ('+gridname+')', ext)
    plot_voltages(Vlst, 'HELM-Z voltage convergence ('+gridname+')', ext=ext)
